import os.path

from fabric.connection import Connection

from . import files


def generate_ssl_dhparams(c: Connection):
    dhparams = "/etc/nginx/ssl/dhparams.pem"
    if not files.exists(c, dhparams):
        d = os.path.dirname(dhparams)
        if not files.exists(c, d):
            c.run(f"mkdir -p {d}")
        c.run(f"openssl dhparam -out {dhparams} 2048", echo=True)



from __future__ import annotations

import io
import re
from shlex import quote

from fabric.connection import Connection
from fabric.transfer import Transfer
from invoke.runners import Result

__all__ = [
    "append",
    "exists",
    "get_file_as_bytes",
    "put_file_from_bytes",
    "require_directory",
    "require_file",
    "is_file",
    "is_dir",
    "get_owner",
    "get_group",
    "get_mode",
]


def get(c: Connection, remote_filename: str, local_filename: str):
    return Transfer(c).get(local=local_filename, remote=remote_filename).local


def put(c: Connection, local_filename: str, remote_filename: str):
    Transfer(c).put(local=local_filename, remote=remote_filename)


def get_file_as_bytes(c: Connection, filename: str) -> bytes:
    store = io.BytesIO()
    Transfer(c).get(remote=filename, local=store, preserve_mode=False)
    return store.getvalue()


def put_file_from_bytes(
    c: Connection, filename: str, contents: bytes, mode: str | None = None, owner: str | None = None
):
    store = io.BytesIO()
    store.write(contents)
    result = Transfer(c).put(local=store, remote=filename, preserve_mode=False)
    remote_path = result.remote
    if mode is not None:
        c.run(f"chmod {mode} {quote(remote_path)}")
    if owner is not None:
        c.run(f"chmod {mode} {quote(remote_path)}")

    return result


def require_directory(c: Connection, path: str, owner: str = "", group: str = "", mode: str = ""):
    """
    Require a directory to exist.
    """

    if not is_dir(c, path):
        c.run(f"mkdir -p {quote(path)}", echo=True)

    _fix_perms(c, path, owner, group, mode)


def require_file(c: Connection, path: str, owner: str = "", group: str = "", mode: str = ""):
    """
    Require a file to exist
    """
    if not is_file(c, path):
        c.run(f"touch {quote(path)}", echo=True)

    _fix_perms(c, path, owner, group, mode)


def _fix_perms(c: Connection, path: str, owner: str, group: str, mode: str):
    # Ensure correct owner
    if (owner and get_owner(c, path) != owner) or (group and get_group(c, path) != group):
        if owner and group:
            c.run(f"chown {owner}:{group} {quote(path)}", echo=True)
        elif owner:
            c.run(f"chown {owner} {quote(path)}", echo=True)
        elif group:
            c.run(f"chgrp {group}:{group} {quote(path)}", echo=True)

    # Ensure correct mode
    if mode and get_mode(c, path).lstrip("0") != mode.lstrip("0"):
        c.run(f"chmod {mode} {quote(path)}", echo=True)


def is_file(c: Connection, path: str):
    """
    Check if a path exists, and is a file.
    """
    return c.run(f"test -f {quote(path)}", warn=True, hide="both").ok


def is_dir(c: Connection, path: str):
    """
    Check if a path exists, and is a directory.
    """
    return c.run(f"test -d {quote(path)}", warn=True, hide="both").ok


def get_owner(c: Connection, path: str):
    """
    Get the owner name of a file or directory.
    """
    result: Result = c.run(f"stat -c %U {quote(path)}", hide="both")
    return result.stdout.strip()


def get_group(c: Connection, path: str):
    """
    Get the group of a file or directory.
    """
    result: Result = c.run(f"stat -c %G {quote(path)}", hide="both")
    return result.stdout.strip()


def get_mode(c: Connection, path: str):
    """
    Get the mode/permissions of a file or directory.
    """
    result: Result = c.run(f"stat -c %a {quote(path)}", hide="both")
    return result.stdout.strip()


def exists(c: Connection, path: str):
    """
    Return True if given path exists on the current remote host.
    """
    result: Result = c.run(f'test -e "$(echo {quote(path)})"', hide="both", warn=True)
    return result.ok


def append(c: Connection, filename: str, text: str | list[str], partial=False, escape=True):
    """
    Append string (or list of strings) ``text`` to ``filename``.

    When a list is given, each string inside is handled independently (but in
    the order given.)

    If ``text`` is already found in ``filename``, the append is not run, and
    None is returned immediately. Otherwise, the given text is appended to the
    end of the given ``filename`` via e.g. ``echo '$text' >> $filename``.

    The test for whether ``text`` already exists defaults to a full line match,
    e.g. ``^<text>$``, as this seems to be the most sensible approach for the
    "append lines to a file" use case. You may override this and force partial
    searching (e.g. ``^<text>``) by specifying ``partial=True``.

    Because ``text`` is single-quoted, single quotes will be transparently
    backslash-escaped. This can be disabled with ``escape=False``.
    """
    # Normalize non-list input to be a list
    lines: list[str]
    if isinstance(text, str):
        lines = [text]
    else:
        lines = text
    for line in lines:
        regex = "^" + _escape_for_regex(line) + ("" if partial else "$")
        if line and exists(c, filename) and contains(c, filename, regex, escape=False):
            continue
        line = line.replace("'", r"'\\''") if escape else line
        c.run(f"echo '{line}' >> {quote(filename)}")


def contains(c: Connection, filename: str, text: str, exact=False, escape=True):
    """
    Return True if ``filename`` contains ``text`` (which may be a regex.)

    By default, this function will consider a partial line match (i.e. where
    ``text`` only makes up part of the line it's on). Specify ``exact=True`` to
    change this behavior so that only a line containing exactly ``text``
    results in a True return value.

    This function leverages ``egrep`` on the remote end (so it may not follow
    Python regular expression syntax perfectly), and skips the usual outer
    ``env.shell`` wrapper that most commands execute with.

    If ``escape`` is False, no extra regular expression related escaping is
    performed (this includes overriding ``exact`` so that no ``^``/``$`` is
    added.)
    """
    if escape:
        text = _escape_for_regex(text)
        if exact:
            text = f"^{text}$"
    egrep_cmd = f'egrep "{text}" "{quote(filename)}"'
    return c.run(egrep_cmd, hide=True, warn=True).ok


def _escape_for_regex(text):
    """Escape ``text`` to allow literal matching using egrep"""
    regex = re.escape(text)
    # Seems like double escaping is needed for \
    regex = regex.replace("\\\\", "\\\\\\")
    # Triple-escaping seems to be required for $ signs
    regex = regex.replace(r"\$", r"\\\$")
    # Whereas single quotes should not be escaped
    regex = regex.replace(r"\'", "'")
    return regex



from fabric.connection import Connection

from . import files


def add_swap(c: Connection, size="1G", swappiness="10"):
    # Needed to compile some things, and for some occassional processes that
    # need a lot of memory.
    if not files.exists(c, "/swapfile"):
        c.run(f"fallocate -l {size} /swapfile", echo=True)
        c.run("chmod 600 /swapfile", echo=True)
        c.run("mkswap /swapfile", echo=True)
        c.run("swapon /swapfile", echo=True)
        files.append(c, "/etc/fstab", "/swapfile   none    swap    sw    0   0\n")

    # Change swappiness so that we actually use the swap
    c.run(f"sysctl vm.swappiness={swappiness}", echo=True)
    files.append(c, "/etc/sysctl.conf", f"vm.swappiness={swappiness}\n")



from fabric.connection import Connection


def install(c: Connection, project_locale):
    locale = project_locale.replace("UTF-8", "utf8")
    if locale not in c.run("locale -a", hide=True).stdout:
        c.run(f"locale-gen {project_locale}", echo=True)
        c.run(f"update-locale {project_locale}", echo=True)



import posixpath
from shlex import quote

from fabric.connection import Connection

from . import files


# Users
def user_exists(c: Connection, user_name: str) -> bool:
    return c.run(f"id {quote(user_name)}", hide="both", echo=False, warn=True).ok


def create_user(
    c: Connection,
    user_name: str,
    group: str = None,
    shell: str = "/bin/bash",
    ssh_public_keys: list[str] = None,
):
    """
    Create a new user and its home directory.


    *ssh_public_keys* can be a list of (local) filenames of public keys that should be
    added to the user's SSH authorized keys

    """

    # Note that we use useradd (and not adduser), as it is the most
    # portable command to create users across various distributions:
    # http://refspecs.linuxbase.org/LSB_4.1.0/LSB-Core-generic/LSB-Core-generic/useradd.html

    args = ["--user-group"]
    if shell:
        args.extend(["-s", shell])
    args.append(user_name)
    args_str = " ".join(quote(arg) for arg in args)
    c.run(f"useradd {args_str}", echo=True)
    files.require_directory(c, home_directory(c, user_name), owner=user_name, group=user_name)

    if ssh_public_keys:
        if isinstance(ssh_public_keys, str):
            ssh_public_keys = [ssh_public_keys]
        add_ssh_public_keys(c, user_name, ssh_public_keys)


def add_ssh_public_keys(c: Connection, user_name: str, filenames: list[str]):
    """
    Add multiple public keys to the user's authorized SSH keys.

    *filenames* must be a list of local filenames of public keys that
    should be added to the user's SSH authorized keys.

    Example::

        import fabtools

        fabtools.user.add_ssh_public_keys('alice', [
            '~/.ssh/id1_rsa.pub',
            '~/.ssh/id2_rsa.pub',
        ])

    """
    ssh_dir = posixpath.join(home_directory(c, user_name), ".ssh")
    files.require_directory(c, ssh_dir, mode="700", owner=user_name)

    authorized_keys_filename = posixpath.join(ssh_dir, "authorized_keys")
    files.require_file(c, authorized_keys_filename, mode="600", owner=user_name)

    for filename in filenames:
        with open(filename) as public_key_file:
            public_key = public_key_file.read().strip()

        if public_key not in get_authorized_keys(c, user_name):
            files.append(c, filename=authorized_keys_filename, text=public_key)


def get_authorized_keys(c: Connection, user_name: str) -> list[str]:
    """
    Get the list of authorized SSH public keys for the user
    """

    ssh_dir = posixpath.join(home_directory(c, user_name), ".ssh")
    authorized_keys_filename = posixpath.join(ssh_dir, "authorized_keys")
    data = files.get_file_as_bytes(c, authorized_keys_filename)
    return [line for line in data.decode("utf-8").splitlines() if line and not line.startswith("#")]


def home_directory(c: Connection, name: str):
    """
    Get the absolute path to the user's home directory
    """
    return c.run("echo ~" + name, hide=True).stdout.strip()



from fabric import Connection


def enable(c: Connection, service: str):
    c.run(f"update-rc.d {service} defaults")
    c.run(f"service {service} start")


def disable(c: Connection, service: str):
    c.run(f"update-rc.d {service} disable")
    c.run(f"service {service} stop")



from __future__ import annotations

import os.path
from dataclasses import dataclass

from fabric.connection import Connection

from . import files
from .files import get_file_as_bytes, put_file_from_bytes


@dataclass
class Template:
    system: bool
    local_path: str
    remote_path: str
    reload_command: str | None = None
    owner: str | None = None
    mode: str | None = None


def upload_template_and_reload(c: Connection, template: Template, context_data: dict):
    """
    Uploads a template only if it has changed, and if so, reload the
    related service.
    """
    local_path = template.local_path
    if not os.path.exists(local_path):
        project_root = os.path.dirname(os.path.abspath(__file__))
        local_path = os.path.join(project_root, local_path)
    remote_path = template.remote_path
    reload_command = template.reload_command
    remote_data = ""
    if files.exists(c, remote_path):
        remote_data = get_file_as_bytes(c, remote_path).decode("utf-8")

    with open(local_path) as f:
        local_data = f.read()
        local_data %= context_data

    def clean(s):
        return s.replace("\n", "").replace("\r", "").strip()

    if clean(remote_data) == clean(local_data):
        return

    put_file_from_bytes(c, remote_path, local_data.encode("utf-8"), mode=template.mode, owner=template.owner)
    if reload_command:
        c.run(reload_command, echo=True)



from fabric.connection import Connection


def disable_root_login_with_password(c: Connection):
    c.run("sed -i 's:RootLogin yes:RootLogin without-password:' /etc/ssh/sshd_config", echo=True)
    c.run("service ssh restart", echo=True)



from collections.abc import Callable
from contextlib import contextmanager
from typing import Any, TypeVar, cast

from fabric import task as fabtask
from fabric.connection import Connection
from invoke import task as invoke_task
from invoke.config import Config
from invoke.context import Context
from makefun import wraps

# -- Generic connection utilities  ---

# Constraints:
# - app user doesn't have sudo rights on server, for tighter security.
#   This means we can't use a single connection and 'sudo'
# - so we do root tasks using separate root user login connection,
#   only where necessary.
# - we want to call both root tasks and normal tasks directly at
#   the top level
# - we want to call root tasks from normal tasks, and via versa.
# - we want to be able to support passing `--hosts` from CLI
#   (this is typically only done when migrating from one host to another)
# - thankfully, we don't need to support multiple hosts.

# So we define `task` and `root_task` decorators to manage this.


@contextmanager
def set_connection_user(connection: Connection, new_user: str) -> Connection:
    """
    Context manager to change the connection to the specified user,
    creating a new connection if needed.
    """
    if connection.user == new_user:
        yield connection
    else:
        new_connection = get_or_create_connection(connection.host, connection.port, new_user, connection.config)
        yield new_connection
        # Stash after first use, so that it will be active
        # when we next come to use it.
        stash_connection(new_connection)


_CACHED_CONNECTIONS: dict[tuple[str, int, str], Connection] = {}


def get_or_create_connection(host: str, port: int, user: str, config: Config):
    """
    Get a connection matching user@host:port and the config, or create
    a new one
    """
    # Config is not hashable, so we can't put it in dict, but we can compare
    # for object equality. So for the common case of not changing config, we
    # can re-use already opened connections.

    cached = _CACHED_CONNECTIONS.get((host, port, user), None)
    if cached is not None and cached.config == config:
        return cached
    return Connection(host=host, port=port, user=user, config=config)


def stash_connection(connection: Connection):
    if connection.is_connected:
        _CACHED_CONNECTIONS[connection.host, connection.port, connection.user] = connection


F = TypeVar("F", bound=Callable[..., Any])


def managed_connection_task(user: str, host: str) -> Callable[..., Callable[[F], F]]:
    """
    Returns a decorator (generator) that is like `@task()` from
    from fabric, but handles connections so that:
    - the host is the specified host by default
    - the task is passed a connection opened with the specified username

    Normally used like:

      myuser_task = managed_connection_task("myuser", "myhost")

      @myuser_task(...)
      def hello(c):
          c.run("echo hello")

    """

    def _generator(*task_args, **task_kwargs) -> Callable[[F], F]:

        # Set the "hosts" string so that top-level calling of the task works
        # (unless `hosts=` was specified at task level)
        if "hosts" not in task_kwargs:
            task_kwargs["hosts"] = [f"{user}@{host}"]

        def _decorator(task_func: F) -> F:
            @wraps(task_func)
            def _wrapper(c, *args, **kwargs):
                # Inside the function, we have to swap out the connection if we
                # get one with the wrong user. This occurs when the task is called
                # from another.
                with set_connection_user(c, user) as new_c:
                    retval = task_func(new_c, *args, **kwargs)
                return retval

            return cast(F, fabtask(*task_args, **task_kwargs)(_wrapper))

        return _decorator

    return _generator


def local_task(*task_args, **task_kwargs):
    """
    Decorator that ensure a task is run only as a local task,
    never using remote connections.
    """

    def decorator(func):
        @invoke_task(*task_args, **task_kwargs)
        @wraps(func)
        def wrapper(c, *args, **kwargs):
            if isinstance(c, Connection):
                # Replace fabric's Connection with Invoke Context.
                # We perhaps need something better here to respect any other config?
                context = Context()
            else:
                context = c
            return func(context, *args, **kwargs)

        return wrapper

    return decorator



from fabric.connection import Connection


def update_upgrade(c: Connection):
    c.run("DEBIAN_FRONTEND=noninteractive apt update -y", echo=True)
    c.run("DEBIAN_FRONTEND=noninteractive apt upgrade -y", echo=True)


def install(c: Connection, packages: list[str]):
    """
    Installs one or more system packages via apt.
    """
    return c.run("DEBIAN_FRONTEND=noninteractive apt-get install -y -q " + " ".join(packages), echo=True)


def remove(c: Connection, packages: list[str]):
    c.run("DEBIAN_FRONTEND=noninteractive apt remove -y " + " ".join(packages), echo=True)



from __future__ import annotations

from dataclasses import dataclass
from shlex import quote
from typing import Protocol

from fabric.connection import Connection
from invoke.runners import Result

# -- Database utilities


@dataclass
class Database:
    name: str
    user: str
    password: str
    port: str
    locale: str


class DbCommand(Protocol):
    def execute(self, c: Connection) -> Result:
        pass


@dataclass
class PsqlCommand:
    sql: str
    db: Database
    run_as_postgres: bool = False
    return_tuples: bool = False

    def execute(self, c: Connection, hide: str | None = "stdout") -> Result:
        with c.cd("/"):  # avoid "could not change directory" warnings
            user = "postgres" if self.run_as_postgres else self.db.user
            db_name = "postgres" if self.run_as_postgres else self.db.name
            extra_args = "-t" if self.return_tuples else ""
            cmd = f"psql -p {self.db.port} -U {user} -d {db_name} {extra_args} -c {quote(self.sql)}"
            pg_env = db_to_pg_envs(self.db)
            if self.run_as_postgres:
                cmd = f"sudo -u postgres {cmd}"
                pg_env.pop("PGUSER")
                pg_env.pop("PGPASSWORD")
            return c.run(cmd, echo=True, hide=hide, env=pg_env)


PG_ENVIRON_MAP = {
    "name": "PGDATABASE",
    "port": "PGPORT",
    "user": "PGUSER",
    "password": "PGPASSWORD",
    # Not in Database class:
    #   "host": "PGHOST",
}


def db_to_pg_envs(db: Database) -> dict:
    """
    Returns the environment variables that postgres CLI tools like psql
    and pg_dump use, as a dict.
    """
    return {env_name: str(getattr(db, attr_name)) for attr_name, env_name in PG_ENVIRON_MAP.items()}


@dataclass
class DbCommandGroup:
    commands: list[DbCommand]

    def execute(self, c: Connection) -> Result:
        for command in self.commands:
            command.execute(c)


# DB commands


def sql_quote(sql: str):
    return sql.replace("'", "''")


def create_db_command(db: Database) -> DbCommand:
    return DbCommandGroup(
        [
            # CREATE DATABASE command has to be in separate call, otherwise get
            #   "CREATE DATABASE cannot run inside a transaction block"
            PsqlCommand(
                sql=(
                    f"CREATE DATABASE {db.name} "
                    f"  TEMPLATE = template0 ENCODING = 'UTF8' LC_CTYPE = '{db.locale}' LC_COLLATE = '{db.locale}';"
                ),
                db=db,
                run_as_postgres=True,
            ),
            PsqlCommand(
                sql=(
                    f"GRANT ALL ON DATABASE {db.name} TO {db.user}; "
                    + f"ALTER USER {db.user} CREATEDB; "
                    + f"GRANT ALL ON SCHEMA public TO {db.user}; "
                    + f"ALTER DATABASE {db.name} OWNER TO {db.user}; "
                ),
                db=db,
                run_as_postgres=True,
            ),
        ]
    )


def create_db(c: Connection, db: Database) -> Result:
    return create_db_command(db).execute(c)


def drop_db_if_exists_command(db: Database) -> DbCommand:
    return PsqlCommand(
        sql=f"DROP DATABASE IF EXISTS {db.name};",
        db=db,
        run_as_postgres=True,
    )


def drop_db_if_exists(c: Connection, db: Database) -> Result:
    return drop_db_if_exists_command(db).execute(c)


@dataclass
class PGRestore:
    db: Database
    filename: str
    clean: bool = False

    def execute(self, c: Connection) -> Result:
        with c.cd("/"):
            db = self.db
            clean = " -c " if self.clean else ""
            cmd = (
                f"PGPASSWORD={db.password} pg_restore -h localhost -p {db.port} -w -O "
                + f"-U {db.user} "
                + f"{clean} "
                + f"-d {db.name} {quote(self.filename)}"
            )

            return c.run(cmd, echo=True)


def restore_db_command(db: Database, filename: str) -> DbCommand:
    return PGRestore(db, filename)


def restore_db(c: Connection, db: Database, filename: str) -> Result:
    return restore_db_command(db, filename).execute(c)


def create_user_command(db: Database, username: str, password: str) -> DbCommand:
    return PsqlCommand(
        f"CREATE USER {sql_quote(username)} WITH PASSWORD '{sql_quote(password)}';",
        db=db,
        run_as_postgres=True,
    )


def create_user(c: Connection, db: Database, username: str, password: str) -> Result:
    return create_user_command(db, username, password).execute(c)


def create_default_user(c: Connection, db: Database) -> Result:
    return create_user(c, db, db.user, db.password)


def check_user_exists_command(db: Database, username: str) -> DbCommand:
    return PsqlCommand(
        f"""SELECT COUNT(*) FROM pg_user WHERE usename='{sql_quote(username)}'; """,
        db,
        # Normally, doesn't technically require postgres user.
        # But we run command when there might not be any other
        # user or DB existing.
        run_as_postgres=True,
        return_tuples=True,
    )


def check_user_exists(c: Connection, db: Database, username: str) -> bool:
    return check_user_exists_command(db, username).execute(c).stdout.strip() == "1"


def check_database_exists_command(db: Database) -> DbCommand:
    return PsqlCommand(
        f"""SELECT COUNT(*) FROM pg_database WHERE datname='{sql_quote(db.name)}'; """,
        db,
        # Normally, doesn't technically require postgres user.
        # But we run command when there might not be any other
        # user or DB existing.
        run_as_postgres=True,
        return_tuples=True,
    )


def check_database_exists(c: Connection, db: Database) -> bool:
    return check_database_exists_command(db).execute(c).stdout.strip() == "1"
